package pku.sei.dependencegraph;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;

import pku.sei.dependencegraph.graph.Graph;
import pku.sei.dependencegraph.graph.INodeType;

import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.ipa.callgraph.AnalysisCache;
import com.ibm.wala.ipa.callgraph.AnalysisOptions;
import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CallGraphBuilder;
import com.ibm.wala.ipa.callgraph.CallGraphBuilderCancelException;
import com.ibm.wala.ipa.callgraph.Entrypoint;
import com.ibm.wala.ipa.callgraph.impl.Util;
import com.ibm.wala.ipa.cha.ClassHierarchy;
import com.ibm.wala.ipa.cha.ClassHierarchyException;
import com.ibm.wala.ipa.cha.IClassHierarchy;
import com.ibm.wala.ipa.slicer.SDG;
import com.ibm.wala.ipa.slicer.SDGSupergraph2;
import com.ibm.wala.ipa.slicer.Slicer.ControlDependenceOptions;
import com.ibm.wala.ipa.slicer.Slicer.DataDependenceOptions;
import com.ibm.wala.ipa.slicer.Statement;
import com.ibm.wala.ipa.slicer.StatementWithInstructionIndex;
import com.ibm.wala.util.config.AnalysisScopeReader;

public class DependenceGraphExtractor {
	private Graph dg;

	private class CallSite {
		private int cgNodeId;
		private int instructionId;
		CallSite(int cgNodeId, int instructionId) {
			this.cgNodeId = cgNodeId;
			this.instructionId = instructionId;
		}
		@Override
		public boolean equals(Object o) {
			if (o instanceof CallSite) {
				CallSite p = (CallSite)o;
				return p.cgNodeId == cgNodeId &&
						p.instructionId == instructionId;
			}
			return false;
		}
		@Override
		public int hashCode() {
			return (cgNodeId * 13537 + instructionId) * 235711;
		}
	}

	private class BraceManager {
		private HashMap<CallSite, Integer> braceIdMapping = new HashMap<CallSite, Integer>();
		private Integer braceCounter = 0;
		int getBraceId(int graphNodeId, int instructionIndex) {
			CallSite callSite = new CallSite(graphNodeId, instructionIndex);
			if (!braceIdMapping.containsKey(callSite)) {
				braceCounter++;
				braceIdMapping.put(callSite, braceCounter);
			}
			return braceIdMapping.get(callSite);
		}
	}

	private int getStatementType(SDGSupergraph2 supergraph, Statement stmt) {
		byte indicator = INodeType.NORMAL_NODE;
		if (supergraph.isCall(stmt))
			indicator = INodeType.CALL_NODE;
		else if (supergraph.isEntry(stmt))
			indicator = INodeType.ENTRY_NODE;
		else if (supergraph.isExit(stmt))
			indicator = INodeType.EXIT_NODE;
		else if (supergraph.isReturn(stmt))
			indicator = INodeType.RETURN_NODE;
		return indicator;
	}
	
	private AnalysisScope createAnalysisScope(String targetClassPath, String exclusionFilePath) {
		AnalysisScope scope = null;
		try {
			scope = AnalysisScopeReader.makeJavaBinaryAnalysisScope(targetClassPath, new File(exclusionFilePath));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./build", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/ant-launcher.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/ant.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/check.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/derby.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/janino.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/javac.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/jcommon-1.0.9.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/jl1.0.jar", new File(exclusionFilePath)));		
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/sunflow.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/Tidy.jar", new File(exclusionFilePath)));
			scope.addToScope(AnalysisScopeReader.makeJavaBinaryAnalysisScope("./SPECjvm2008/lib/xom-1.1.jar", new File(exclusionFilePath)));
		} catch (IOException e) {
			e.printStackTrace();
		}
		return scope;
	}
	
	private IClassHierarchy makeClassHierarchy(AnalysisScope scope) {
		IClassHierarchy cha = null;
		try {
			cha = ClassHierarchy.make(scope);
		} catch (ClassHierarchyException e) {
			e.printStackTrace();
		}
		return cha;
	}
	
	private boolean isVirtual(IMethod method) {
		return method.isProtected() ||
			!method.isPrivate() && !method.isInit() && !method.isStatic();
	}
	
	private boolean isClientCode(IMethod method, String clientCodeSignature) {
		return method.toString().contains(clientCodeSignature);
	}
	
	private void collectEdges(SDGSupergraph2 supergraph, String clientCodeSignature) {
		Iterator<Statement> srcStmtIt = supergraph.iterator();
		BraceManager bm = new BraceManager();
		while (srcStmtIt.hasNext()) {
			Statement srcStmt = srcStmtIt.next();
			int srcId = supergraph.getNumber(srcStmt);
			CGNode srcCGNode = srcStmt.getNode();
			// TODO isVirtual: call/return(?) entry/exit(done)
			dg.initNode(srcId, 
					getStatementType(supergraph, srcStmt), 
					srcCGNode.getGraphNodeId(), 
					isVirtual(srcCGNode.getMethod()),
					isClientCode(srcCGNode.getMethod(), clientCodeSignature),
					srcStmt
					);
			Iterator<Statement> tgtStmtIt = supergraph.getSuccNodes(srcStmt);
			while (tgtStmtIt.hasNext()) {
				Statement tgtStmt = tgtStmtIt.next();
				int tgtId = supergraph.getNumber(tgtStmt);
				if (supergraph.isCall(srcStmt) && supergraph.isEntry(tgtStmt)) {
					int graphNodeId = srcCGNode.getGraphNodeId();
					int instructionIndex = ((StatementWithInstructionIndex)srcStmt).getInstructionIndex();
					int braceId = bm.getBraceId(graphNodeId, instructionIndex);
					dg.addEdge(srcId, tgtId, braceId);
				}
				else if (supergraph.isExit(srcStmt) && supergraph.isReturn(tgtStmt)) {
					CGNode tgtCGNode = tgtStmt.getNode();
					int graphNodeId = tgtCGNode.getGraphNodeId();
					int instructionIndex = ((StatementWithInstructionIndex)tgtStmt).getInstructionIndex();
					int braceId = bm.getBraceId(graphNodeId, instructionIndex);					
					dg.addEdge(srcId, tgtId, -braceId);
				}
				else { // Normal Edge
					dg.addEdge(srcId, tgtId, 0);
				}
			}
		}
	}
	
	public Graph run(String walaExclusionFilePath, 
			String targetClassPathOrJar, String targetMainClass, 
			String clientCodeSignature){
		try {
			AnalysisScope scope = null;
			scope = createAnalysisScope(targetClassPathOrJar, walaExclusionFilePath);
			IClassHierarchy cha = makeClassHierarchy(scope);
			Iterable<Entrypoint> entrypoints = null;
			if (!targetMainClass.equals("null"))
				entrypoints = Util.makeMainEntrypoints(scope, cha, WALAUtils.getWALAClassName(targetMainClass));
			else
				entrypoints = Util.makeMainEntrypoints(scope, cha);
			
			AnalysisOptions options = new AnalysisOptions(scope, entrypoints);
			options.setHandleStaticInit(true);
			
			CallGraphBuilder builder = Util.makeVanillaZeroOneCFABuilder(options, new AnalysisCache(), cha, scope);
			CallGraph cg = builder.makeCallGraph(options, null);
			
			SDG sdg = new SDG(cg, builder.getPointerAnalysis(), DataDependenceOptions.NO_BASE_NO_HEAP_NO_EXCEPTIONS, ControlDependenceOptions.NONE);
			
			SDGSupergraph2 supergraph = new SDGSupergraph2(sdg, false);
			
			dg = new Graph(supergraph.getNumberOfNodes());
			collectEdges(supergraph, clientCodeSignature);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (CallGraphBuilderCancelException e) {
			e.printStackTrace();
		}
		return dg;
	}
}
