package pku.sei.dependencegraph.graph;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;

import com.ibm.wala.ipa.slicer.Statement;

public class Graph {
	class Edge {
		int src, tgt, edgeType;
		Edge(int src, int tgt, int edgeType) {
			this.src = src;
			this.tgt = tgt;
			this.edgeType = edgeType;
		}
	}
	private ArrayList<Integer> nodeType;
	private ArrayList<Integer> nodeInMethod;
	// nodeIsVirtual: "invokevirtual" or "invokeinterface"
	// nodeIsClientCode
	private ArrayList<Boolean> nodeIsVirtual;  
	private ArrayList<Boolean> nodeIsClientCode;
	private ArrayList<Boolean> nodeIsKey;
	private ArrayList<Boolean> nodeIsBoundary;
	private ArrayList<ArrayList<Edge>> forwardEdges;
	private ArrayList<ArrayList<Edge>> backwardEdges;
	private Iterator<Edge> edgeIterator;
	private Edge currentEdge;
	private ArrayList<Boolean> nodeHasCallInfo;
	private int n;
	private ArrayList<Statement> statements;
	
	public Graph(int n) {
		this.n = n;
		nodeType = new ArrayList<Integer>(n);
		nodeInMethod = new ArrayList<Integer>(n);
		nodeIsVirtual = new ArrayList<Boolean>(n);
		nodeIsKey = new ArrayList<Boolean>(n);
		nodeIsBoundary = new ArrayList<Boolean>(n);
		nodeIsClientCode = new ArrayList<Boolean>(n);
		forwardEdges = new ArrayList<ArrayList<Edge>>(n);
		backwardEdges = new ArrayList<ArrayList<Edge>>(n);
		statements = new ArrayList<Statement>(n);
		nodeHasCallInfo = new ArrayList<Boolean>(n);
		for (int i = 0; i < n; i++) {
			nodeType.add(i, INodeType.NORMAL_NODE);
			nodeInMethod.add(i, 0);
			nodeIsVirtual.add(i, false);
			nodeIsClientCode.add(i, false);
			nodeIsKey.add(i, false);
			nodeIsBoundary.add(i, false);
			forwardEdges.add(i, new ArrayList<Edge>());
			backwardEdges.add(i, new ArrayList<Edge>());
			statements.add(i, null);
		}
	}
	public int getN() {
		return n;
	}
	public void initNode(int node, int t, int methodId, boolean isVirtual, boolean isClientCode, Statement stmt) {
		nodeType.add(node, t);
		nodeInMethod.add(node, methodId);
		nodeIsVirtual.add(node, isVirtual);
		nodeIsClientCode.add(node, isClientCode);
		statements.add(node, stmt);
	}
	public int getNodeType(int node) {
		return nodeType.get(node);
	}
	public boolean getNodeIsVirtual(int node) {
		return nodeIsVirtual.get(node);
	}
	public boolean setNodeIsVirtual(int node, boolean element) {
		return nodeIsVirtual.set(node, element);
	}
	public boolean getNodeIsClientCode(int node) {
		return nodeIsClientCode.get(node);
	}
	public boolean getNodeIsBoundary(int node) {
		return nodeIsBoundary.get(node);
	}
	public int getMethodId(int node) {
		return nodeInMethod.get(node);
	}
	public Statement getStatement(int node) {
		return statements.get(node);
	}
	public boolean getNodeHasCallInfo(int node) {
		return nodeHasCallInfo.get(node);
	}
	public boolean getNodeIsKey(int node) {
		return nodeIsKey.get(node);
	}
	public void addEdge(int src, int tgt, int edgeType) {
		edgeType = IPathType_1.convert2PathType(edgeType);
		Edge e = new Edge(src, tgt, edgeType);
		forwardEdges.get(src).add(e);
		backwardEdges.get(tgt).add(e);
	}
	public void initializeForwardEdgeIterator(int src) {
		edgeIterator = forwardEdges.get(src).iterator();
	}
	public void initializeBackwardEdgeIterator(int tgt) {
		edgeIterator = backwardEdges.get(tgt).iterator();
	}
	public boolean hasNextEdge() {
		if (edgeIterator.hasNext()) {
			currentEdge = edgeIterator.next();
			return true;
		}
		else
			return false;
	}
	public int getForwardEdgeTarget() {
		return currentEdge.tgt; 
	}
	public int getBackwardEdgeSource() {
		return currentEdge.src; 
	}
	public int getEdgeType() {
		return currentEdge.edgeType; 
	}
	public int getForwardEdgeNumber(int node) {
		return forwardEdges.get(node).size();
	}
	public int getBackwardEdgeNumber(int node) {
		return backwardEdges.get(node).size();
	}
	public void propagateCallInfoAndLabelKeyNodes() {
		nodeHasCallInfo = new ArrayList<Boolean>(n);
		Queue<Integer> bfsQueue = new LinkedList<Integer>();
		for (int i = 0; i < n; i++) {
			if (nodeType.get(i) == INodeType.CALL_NODE) {
				nodeHasCallInfo.add(i, true);
				bfsQueue.offer(i);
			}
			else
				nodeHasCallInfo.add(i, false);
		}
		while (bfsQueue.size() > 0) {
			int u = bfsQueue.poll();
			initializeBackwardEdgeIterator(u);
			while (hasNextEdge()) {
				int v = getBackwardEdgeSource();
				if (!nodeHasCallInfo.get(v)) {
					nodeHasCallInfo.add(v, true);
					if (nodeType.get(v) == INodeType.NORMAL_NODE) {
						bfsQueue.offer(v);
					}
				}
			}
		}
		// Date: 2014-6-4
		// Label keynode-RETURN&ENTRY&EXIT 
		Set<Integer> keyMethodSet = new TreeSet<Integer>();
		for (int i = 0; i < n; i++) {
			if (this.getNodeType(i) == INodeType.RETURN_NODE && 
					this.getNodeHasCallInfo(i)) {
				keyMethodSet.add(this.getMethodId(i));
				nodeIsKey.set(i, true); // KeyNode:RETURN
			}
		}
		for (int i = 0; i < n; i++) {
			if (this.getNodeType(i) == INodeType.ENTRY_NODE) {
				if (keyMethodSet.contains(this.getMethodId(i)))
					nodeIsKey.set(i, true); // KeyNode:ENTRY
			}
			else if (this.getNodeType(i) == INodeType.EXIT_NODE) {
				if (keyMethodSet.contains(this.getMethodId(i)))
					nodeIsKey.set(i, true); // KeyNode:EXIT
			}
		}
		
	}
	
	public void labelBoundary() {
		for (int i = 0; i < n; i++) {
			if (this.getNodeIsClientCode(i)) {
				this.initializeForwardEdgeIterator(i);
				while (this.hasNextEdge()) {
					int tgt = this.getForwardEdgeTarget();
					if (!this.getNodeIsClientCode(tgt)) {
						nodeIsKey.set(tgt, true);
						nodeIsBoundary.set(tgt, true);
					}
				}
				this.initializeBackwardEdgeIterator(i);
				while (this.hasNextEdge()) {
					int src = this.getBackwardEdgeSource();
					if (!this.getNodeIsClientCode(src)) {
						nodeIsKey.set(src, true);
						nodeIsBoundary.set(src, true);
					}
				}
			}
		}
	}
}
